
/* Chrysalide - Outil d'analyse de fichiers binaires
 * header.c - constitution des deux types d'en-têtes JDWP
 *
 * Copyright (C) 2010-2012 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "header.h"


#include <string.h>


#include "../../../common/endianness.h"



/******************************************************************************
*                                                                             *
*  Paramètres  : header  = en-tête logicielle au format local à constituer.   *
*                blob    = en-tête en gros boutiste du paquet à constituer.   *
*                length  = taille totale du paquet.                           *
*                set     = jeu de commandes de la requête.                    *
*                command = commande proprement dite.                          *
*                                                                             *
*  Description : Définit une en-tête de requête au format JDWP.               *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void set_jdwp_request_header(jdwp_header *header, bin_t *blob, uint32_t length, uint8_t set, uint8_t command)
{
    off_t pos;                              /* Tête d'écriture             */

    pos = 0;

    length += sizeof(jdwp_header);

    /* Encodage local */

    header->length = length;
    header->id = 1;
    header->flags = JDWP_FLAGS_NONE;

    header->set = set;
    header->command = command;

    /* Encodage gros boutiste */

    write_u32(&length, blob, &pos, sizeof(jdwp_header), SRE_BIG);
    write_u32((uint32_t []) { 1 }, blob, &pos, sizeof(jdwp_header), SRE_BIG);
    write_u8((uint8_t []) { JDWP_FLAGS_NONE }, blob, &pos, sizeof(jdwp_header), SRE_BIG);

    write_u8(&set, blob, &pos, sizeof(jdwp_header), SRE_BIG);
    write_u8(&command, blob, &pos, sizeof(jdwp_header), SRE_BIG);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : header = en-tête logicielle au format local à constituer.    *
*                blob   = en-tête en gros boutiste du paquet à constituer.    *
*                length = taille totale du paquet.                            *
*                lastid = jeton du paquet à l'origine du besoin de réponse.   *
*                error  = éventuelle indication d'erreur.                     *
*                                                                             *
*  Description : Définit une en-tête de réponse au format JDWP.               *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void set_jdwp_reply_header(jdwp_header *header, bin_t *blob, uint32_t length, uint32_t lastid, uint16_t error)
{
    off_t pos;                              /* Tête d'écriture             */

    pos = 0;

    length += sizeof(jdwp_header);

    /* Encodage local */

    header->length = length;
    header->id = 1;
    header->flags = JDWP_FLAGS_REPLY;

    header->error = error;

    /* Encodage gros boutiste */

    write_u32(&length, blob, &pos, sizeof(jdwp_header), SRE_BIG);
    write_u32((uint32_t []) { 1 }, blob, &pos, sizeof(jdwp_header), SRE_BIG);
    write_u8((uint8_t []) { JDWP_FLAGS_REPLY }, blob, &pos, sizeof(jdwp_header), SRE_BIG);

    write_u16(&error, blob, &pos, sizeof(jdwp_header), SRE_BIG);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : header  = en-tête logicielle au format local à constituer.   *
*                blob    = en-tête en gros boutiste du paquet à constituer.   *
*                                                                             *
*  Description : Enregistre au format binaire la valeur actuelle de la taille.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void update_jdwp_header_length(const jdwp_header *header, bin_t *blob)
{
    off_t pos;                              /* Tête d'écriture             */

    pos = 0;

    write_u32(&header->length, blob, &pos, sizeof(jdwp_header), SRE_BIG);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : blob   = flux de données à analyser.                         *
*                header = en-tête de paquet JDWP reconstituée. [OUT]          *
*                                                                             *
*  Description : Lit une en-tête de paquet au format JDWP.                    *
*                                                                             *
*  Retour      : Bilan de l'opération : true en cas de succès, false sinon.   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool get_jdwp_header(const bin_t *blob, jdwp_header *header)
{
    off_t pos;                              /* Tête de lecture             */
    off_t len;                              /* Taille standard d'en-tête   */

    pos = 0;
    len = sizeof(jdwp_header);

    if (!read_u32(&header->length, blob, &pos, len, SRE_BIG))
        return false;

    if (!read_u32(&header->id, blob, &pos, len, SRE_BIG))
        return false;

    if (!read_u8(&header->flags, blob, &pos, len))
        return false;

    /* Réponse ? */
    if (header->flags & JDWP_FLAGS_REPLY)
    {
        if (!read_u16(&header->error, blob, &pos, len, SRE_BIG))
            return false;
    }

    /* Requête ! */
    else
    {
        if (!read_u8(&header->set, blob, &pos, len))
            return false;

        if (!read_u8(&header->command, blob, &pos, len))
            return false;

    }

    return true;

}
